#include <ctype.h>
#include <getopt.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "parsing.h"
#include "scan.h"

void free_options(struct option_lst *options)
{
	struct option_lst *it = options;
	while (it) {
		struct option_lst *tmp = it;
		it = it->next;
		free(tmp);
	}
}

bool option_isset(const struct option_lst *options, e_flag flag)
{
	if (!options)
		return false;
	for (const struct option_lst *it = options; it; it = it->next)
		if (it->flag == flag)
			return true;
	return false;
}

char *get_option_arg(const struct option_lst *options, e_flag flag)
{
	if (!options)
		return NULL;
	for (const struct option_lst *it = options; it; it = it->next)
		if (it->flag == flag)
			return it->arg;
	return NULL;
}

uint8_t parse_type(char *arg)
{
	if (!arg)
		return SCAN_ALL;
	uint8_t type = 0;
	char *current_arg = arg;
	while (current_arg[0] != '\0')
	{
		if (strncmp(current_arg, "NULL", 4) == 0)
			type |= SCAN_NULL;
		else if (strncmp(current_arg, "SYN", 3) == 0)
			type |= SCAN_SYN;
		else if (strncmp(current_arg, "ACK", 3) == 0)
			type |= SCAN_ACK;
		else if (strncmp(current_arg, "FIN", 3) == 0)
			type |= SCAN_FIN;
		else if (strncmp(current_arg, "XMAS", 4) == 0)
			type |= SCAN_XMAS;
		else if (strncmp(current_arg, "UDP", 5) == 0)
			type |= SCAN_UDP;
		else
		{
			dprintf(2, "ft_nmap: invalid argument to --scan: '%s'\n", arg);
			return -1;
		}
		current_arg += strcspn(current_arg, ",");
		if (current_arg[0] == ',')
			current_arg++;
	}
	return type;
}

int parse_ports(const char *arg, uint16_t *p_start, uint16_t *p_end)
{
	size_t i = 0;
	if (!arg)
		return 0;
	for (; isdigit(arg[i]); i++)
		;
	int start = atoi(arg);
	if (arg[i] == '\0') {
		*p_start = start;
		*p_end = start;
		return 0;
	}
	if (arg[i++] != '-')
		goto error;
	int end = atoi(arg + i);
	for (; isdigit(arg[i]); i++)
		;
	if (arg[i])
		goto error;
	if (start >= end)
		goto error;
	if (end > UINT16_MAX) {
		dprintf(2,
		        "ft_nmap: invalid argument: '%s': out of range: 0 "
		        "<= value <= %d\n",
		        arg, UINT16_MAX);
		return -1;
	}
	if (start < 0 || end < 0) {
		dprintf(2, "ft_nmap: invalid port value: %d\n", start);
		return -1;
	}
	*p_start = start;
	*p_end = end;
	return 0;
error:
	dprintf(2, "ft_nmap: port: invalid argument: '%s'\n", arg);
	return -1;
}

static int check_arg(e_flag flag, const char *arg)
{
	switch (flag) {
	case FL_SPEEDUP: {
		for (size_t i = 0; arg[i]; i++) {
			if (!isdigit(arg[i])) {
				dprintf(2, "ft_nmap: invalid argument: %s\n",
				        arg);
				return -1;
			}
		}
		int value = atoi(arg);
		if (value < 0 || value > 250) {
			dprintf(2,
			        "ft_nmap: invalid argument: '%s': out "
			        "of range: 0 "
			        "<= value <= 250\n",
			        arg);
			return -1;
		}
		break;
	}
	case FL_TTL: {
		int nb = atoi(arg);
		if (nb == 0) {
			dprintf(2, "ft_nmap: cannot set unicast time-to-live: "
			           "Invalid argument\n");
			return -1;
		} else if (nb < 0 || nb > 255) {
			dprintf(2,
			        "ft_nmap: invalid argument: '%s': out "
			        "of range: 0 "
			        "<= value <= 255\n",
			        arg);
			return -1;
		}
		break;
	}
	case FL_MAXRETRIES: {
		int nb = atoi(arg);
		if (nb == 0) {
			dprintf(2, "ft_nmap: cannot set max_retries to 0\n");
			return -1;
		} else if (nb < 0 || nb > 255) {
			dprintf(2,
			        "ft_nmap: invalid argument: '%s': out "
			        "of range: 0 "
			        "<= value <= 255\n",
			        arg);
			return -1;
		}
		break;
	}
	default:
		break;
	}
	return 0;
}

static int add_option(struct option_lst **head, e_flag flag, char *arg)
{
	struct option_lst *new_option = malloc(sizeof(struct option_lst));
	if (!new_option) {
		dprintf(2, "ft_nmap: allocation of option failed\n");
		for (struct option_lst *it = *head; it;) {
			struct option_lst *tmp = it;
			it = it->next;
			free(tmp);
		}
		return -1;
	}

	new_option->flag = flag;
	new_option->arg = arg;
	new_option->next = NULL;

	if (*head == NULL) {
		*head = new_option;
		return 0;
	}
	for (struct option_lst *it = *head; it; it = it->next) {
		if (it->next == NULL) {
			it->next = new_option;
			break;
		}
	}
	return 0;
}

struct option_lst *parse_options(int ac, char *const *av)
{
	struct option_lst *head = NULL;
	const struct option options[] = {
	    {"help", no_argument, 0, 0},
	    {"ports", required_argument, 0, 0},
	    {"ip", required_argument, 0, 0},
	    {"file", required_argument, 0, 0},
	    {"speedup", required_argument, 0, 0},
	    {"scan", required_argument, 0, 0},
	    {"max_retries", required_argument, 0, 0},
	    {"ttl", required_argument, 0, 0},
	};

	int c;
	while (1) {
		int option_index = 0;
		c = getopt_long(ac, av, "F", options, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case 'F':
			add_option(&head, FL_FAST, optarg);
			continue;
		case '?':
			return NULL;
		default:
			break;
		}
		if (check_arg(option_index, optarg) < 0)
			return NULL;
		if (add_option(&head, option_index, optarg) < 0)
			return NULL;
	}
	return head;
}

int parsing(struct scan *general, const struct option_lst *options)
{
	general->port_start = 1;
	general->port_end = 1024;
	if (option_isset(options, FL_FAST))
		general->port_end = 128;
	if (parse_ports(get_option_arg(options, FL_PORTS), &general->port_start,
	                &general->port_end))
		return -1;
	if (general->port_start <= 0 || general->port_end <= 0) {
		dprintf(2, "ft_nmap: invalid port value: %d\n",
		        general->port_start <= 0 ? general->port_start
		                                 : general->port_end);
		return -1;
	}
	general->type = parse_type(get_option_arg(options, FL_TYPE));
	if (general->type == (uint8_t) -1)
		return -1;
	const char *max_retries = get_option_arg(options, FL_MAXRETRIES);
	general->max_retries = max_retries ? atoi(max_retries) : 1;
	const char *ttl = get_option_arg(options, FL_TTL);
	general->ttl = ttl ? atoi(ttl) : 48;
	const char *nb_threads_str = get_option_arg(options, FL_SPEEDUP);
	uint8_t nb_threads = nb_threads_str ? atoi(nb_threads_str) : 0;
	if (general->port_end - general->port_start + 1 < nb_threads) {
		dprintf(2, "ft_nmap: number of threads to use must be "
		           "superior "
		           "or equals to the ports range\n");
		return -1;
	}
	return 0;
}
